import logging
from .route import create_route

logger = logging.getLogger(__name__)

class Namespace:
    def __init__(self, name):
        self.name = name
        self.models = []
        self.routes = []
    
    
    def model(self, cli=None, http=None):
        def declare_model(Model):
            logger.debug('Add model {} to namespace {}'.format(Model.__name__, self.name))
            self.models.append(Model)
        return declare_model
    

    def method(self, cli=None, http=None, method='get'):
        def declare_method(func):
            """
            Declare method func from the model defined by its class
            """
            logger.debug('declaring method %s'%func)

            if http is not None:
                # declare a new route in api
                print('/'+http)
                self.routes.append(
                    create_route(path='/'+http, callback=func, method=method))
        return declare_method

    def cli(self):
        pass



import apicli.fields

class Model:
    """Base object that can be serialized/deserialized
    """

    def load(self):
        raise NotImplementedError

    def dump(self):
        raise NotImplementedError

    def __str__(self):
        raise NotImplementedError

    @classmethod
    def openapi_definition(cls):
        # list model fields
        properties = [prop for prop in dir(cls) if isinstance(getattr(cls, prop), apicli.fields.Field)]

        return {
            'type': 'object',
            'properties': {
                prop: getattr(cls, prop).specs() for prop in properties
            }
        }



from apicli.model import Model
from apicli.namespace import Namespace
from apicli.fields import *
from apicli.api import API

import logging

logging.basicConfig(
    format='%(asctime)s %(levelname)s\t%(name)-25s %(message)s',
    level=logging.DEBUG
)

logger = logging.getLogger('apicli')



import logging
import tornado.web

from apicli.defaults import SERVER_HOST, SERVER_PORT, DEBUG_MODE


logger = logging.getLogger(__name__)


class ApiSpecs(tornado.web.RequestHandler):
    title = None
    description = None
    version = None
    address = None
    definitions = []

    @classmethod
    def set_address(cls, host, port):
        cls.address = 'http://{}{}/'.format(host, ':' + str(port) if port!=80 else '')

    @classmethod
    def set_models(cls, models):
        cls.definitions = {model.__name__: model.openapi_definition() for model in models}

    def get(self):
        self.set_header("Access-Control-Allow-Origin", "*")
        self.write(self.json())

    def json(self):
        import json
        return json.dumps({
            "swagger": "2.0",
            "host": self.address,
            "schemes": ["https", "http"],
            "info": {
                "description": self.description,
                "version": self.version,
                "title": self.title
            },
            "definitions": self.definitions
        }, indent=4)


class API:
    def __init__(self, *, title, description, version='0.0.1'):
        self.specs = ApiSpecs

        self.specs.title = title
        self.specs.description = description
        self.specs.version = version
        self.specs.address = None

    def add_namespace(self, namespace):
        self.namespace = namespace

    def run(self, host=SERVER_HOST, port=SERVER_PORT, debug=DEBUG_MODE):
        # configure api specs with bind address
        self.specs.set_address(host, port)
        self.specs.set_models(self.namespace.models)

        server = tornado.web.Application([
            (r'/swagger.json', ApiSpecs)
        ])
        server.listen(port=port, address=host)

        if DEBUG_MODE:
            logger.warning('Debug mode enabled')

        # log server address and start
        logger.info(
            'Tornado server listening on {} {}'.format(self.specs.address, 'in debug mode'*DEBUG_MODE))
        tornado.ioloop.IOLoop.current().start()
        


import tornado.web

def create_route(path:str, callback, method='get'):
    name = method.title() + 'Handler' + path.title().replace('/', '')
    print(name)
    return type(
        name, 
        (tornado.web.RequestHandler,), 
        {method: callback})




class Field:
    def __init__(self, *, type, format=None, example=None, required=False):
        self.required = required
        self.type = type
        self.format = format
        self.example = example

    def specs(self):
        _specs = {'type': self.type}
        if self.format:
            _specs['format'] = self.format
        if self.example:
            _specs['example'] = self.example
        return _specs


class String(Field):
    def __init__(self, required=False, format=None, example=None):
        super().__init__(required=required, type='string', format=format, example=example)


class Email(String):
    def __init__(self, required=False, format='email', example='toto@example.com'):
        super().__init__(required=required, format=format, example=example)


class Password(Field):
    def __init__(self, required=False, default=None, format=None, example=None):
        super().__init__(required=required, type='string', format='string', example=example)
